"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[2346],{2346:function(e,t,i){i.r(t),i.d(t,{default:function(){return b}});i(8309);var n=i(6252),s=i(3577),a={class:"camera component-row"},r={class:"feed-container",ref:"container"},c=["textContent"],l=["src"],o={key:2},u=["src"],p={class:"controls"},f={key:0,class:"fa fa-play"},d={key:1,class:"fa fa-pause"};function g(e,t,i,g,v,m){return(0,n.wg)(),(0,n.iD)("div",a,[(0,n._)("div",r,[v.visible?(0,n.kq)("",!0):((0,n.wg)(),(0,n.iD)("div",{key:0,class:"no-content",textContent:(0,s.zw)(i.name)},null,8,c)),v.visible&&"image"===i.type?((0,n.wg)(),(0,n.iD)("img",{key:1,alt:"Camera feed",src:m.imgUrl},null,8,l)):v.visible&&"video"===i.type?((0,n.wg)(),(0,n.iD)("video",o,[(0,n._)("source",{src:i.src},null,8,u)])):(0,n.kq)("",!0)],512),(0,n._)("div",p,[(0,n._)("button",{class:"toggle-btn",onClick:t[0]||(t[0]=function(e){return v.visible=!v.visible})},[v.visible?((0,n.wg)(),(0,n.iD)("i",d)):((0,n.wg)(),(0,n.iD)("i",f))])])])}i(1539),i(9714);var v=i(2628),m={name:"Camera",mixins:[v.Z],props:{src:{type:String,required:!0},type:{type:String,default:"image"},name:{type:String}},computed:{imgUrl:function(){if("image"===this.type)return this.src+(this.src.indexOf("?")>0?"&":"?")+"_t="+(new Date).getTime().toString()}},data:function(){return{visible:!1}}},y=i(3744);const k=(0,y.Z)(m,[["render",g],["__scopeId","data-v-0fad5251"]]);var b=k}}]);
//# sourceMappingURL=2346-legacy.cc7ee826.js.map