"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[4826],{4826:function(n,t,e){e.r(t),e.d(t,{default:function(){return l}});var o=e(6252);const i={class:"plugin"};function s(n,t,e,s,p,a){const c=(0,o.up)("Loading");return(0,o.wg)(),(0,o.iD)("div",i,[p.loading?((0,o.wg)(),(0,o.j4)(c,{key:0})):p.component?((0,o.wg)(),(0,o.j4)((0,o.LL)(p.component),{key:1,config:p.config},null,8,["config"])):(0,o.kq)("",!0)])}var p=e(2628),a=e(1232),c={name:"Plugin",components:{Loading:a.Z},mixins:[p.Z],props:{pluginName:{type:String,required:!0}},data(){return{loading:!1,component:null,config:{}}},computed:{componentName(){return this.pluginName.split(".").map((n=>n[0].toUpperCase()+n.slice(1))).join("")}},methods:{refresh:async function(){this.loading=!0;try{this.component=(0,o.RC)((()=>e(3379)(`./${this.componentName}/Index`))),this.$options.components[this.componentName]=this.component,this.config=(await this.request("config.get_plugins"))?.[this.pluginName]||{}}finally{this.loading=!1}}},mounted:function(){this.refresh()}},u=e(3744);const r=(0,u.Z)(c,[["render",s],["__scopeId","data-v-3565b88b"]]);var l=r}}]);
//# sourceMappingURL=4826.5f8aa5e8.js.map