"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[2466],{2466:function(t,a,e){e.r(a),e.d(a,{default:function(){return d}});var n=e(6252);function l(t,a,e,l,i,s){const u=(0,n.up)("Panel");return(0,n.wg)(),(0,n.j4)(u,{"plugin-name":"tts"})}var i=e(8825),s={name:"Tts",components:{Panel:i.Z}},u=e(3744);const r=(0,u.Z)(s,[["render",l]]);var d=r},8825:function(t,a,e){e.d(a,{Z:function(){return y}});var n=e(6252),l=e(9963);const i=t=>((0,n.dD)("data-v-a248454a"),t=t(),(0,n.Cn)(),t),s={class:"tts-container"},u={class:"field text-container"},r=["disabled"],d={class:"field lang-container"},c=["disabled"],o={class:"field buttons"},p=["disabled"],g=i((()=>(0,n._)("i",{class:"fa fa-volume-up"},null,-1))),f=[g];function v(t,a,e,i,g,v){return(0,n.wg)(),(0,n.iD)("div",s,[(0,n._)("form",{onSubmit:a[0]||(a[0]=(0,l.iM)(((...t)=>v.talk&&v.talk(...t)),["prevent"]))},[(0,n._)("div",u,[(0,n._)("label",null,[(0,n._)("input",{type:"text",name:"text",placeholder:"Text to say",disabled:g.talking},null,8,r)])]),(0,n._)("div",d,[(0,n._)("label",null,[(0,n._)("input",{type:"text",name:"language",placeholder:"Language code",disabled:g.talking},null,8,c)])]),(0,n._)("div",o,[(0,n._)("button",{type:"submit",disabled:g.talking},f,8,p)])],32)])}var m=e(2628),b={name:"Panel",mixins:[m.Z],props:{pluginName:{type:String,required:!0}},data(){return{talking:!1}},methods:{async talk(t){const a=[...t.target.querySelectorAll("input")].reduce(((t,a)=>(a.value.length&&(t[a.name]=a.value),t)),{});this.talking=!0;try{await this.request(`${this.pluginName}.say`,a)}finally{this.talking=!1}}}},h=e(3744);const k=(0,h.Z)(b,[["render",v],["__scopeId","data-v-a248454a"]]);var y=k}}]);
//# sourceMappingURL=2466.575a0b76.js.map