"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[6509],{6509:function(e,a,n){n.r(a),n.d(a,{default:function(){return s}});var u=n(6252);function r(e,a,n,r,t,p){var i=(0,u.up)("Media");return(0,u.wg)(),(0,u.j4)(i,{"plugin-name":"media.mplayer"})}var t=n(4196),p={name:"MediaMplayer",components:{Media:t["default"]}},i=n(3744);const l=(0,i.Z)(p,[["render",r]]);var s=l}}]);
//# sourceMappingURL=6509-legacy.28c562ac.js.map