"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[2820],{2820:function(e,n,r){r.r(n),r.d(n,{default:function(){return d}});var t=r(6252);function a(e,n,r,a,o,i){var u=(0,t.up)("Panel");return(0,t.wg)(),(0,t.j4)(u,{"plugin-name":"rtorrent"})}var o=r(6899),i={name:"Torrent",components:{Panel:o.Z}},u=r(3744);const c=(0,u.Z)(i,[["render",a]]);var d=c},6899:function(e,n,r){r.d(n,{Z:function(){return U}});var t=r(6252),a={class:"torrent-container"},o={class:"header-container"},i={class:"view-container"};function u(e,n,r,u,c,d){var s=(0,t.up)("Header"),l=(0,t.up)("TorrentView");return(0,t.wg)(),(0,t.iD)("div",a,[(0,t._)("div",o,[(0,t.Wm)(s,{onTorrentAdd:n[0]||(n[0]=function(e){return d.download(e)})})]),(0,t._)("div",i,[(0,t.Wm)(l,{"plugin-name":r.pluginName},null,8,["plugin-name"])])])}var c=r(8534),d=(r(5666),r(3577)),s=r(9963),l={class:"row"},p={class:"col-s-12 col-m-9 col-l-7 left side"},m={class:"search-box"};function f(e,n,r,a,o,i){return(0,t.wg)(),(0,t.iD)("div",{class:(0,d.C_)(["header",{"with-filter":e.filterVisible}])},[(0,t._)("div",l,[(0,t._)("div",p,[(0,t._)("form",{onSubmit:n[1]||(n[1]=(0,s.iM)((function(n){return e.$emit("torrent-add",o.torrentURL)}),["prevent"]))},[(0,t._)("label",m,[(0,t.wy)((0,t._)("input",{type:"search",placeholder:"Add torrent URL","onUpdate:modelValue":n[0]||(n[0]=function(e){return o.torrentURL=e})},null,512),[[s.nr,o.torrentURL]])])],32)])])],2)}var v={name:"Header",emits:["torrent-add"],data:function(){return{torrentURL:""}}},w=r(3744);const h=(0,w.Z)(v,[["render",f],["__scopeId","data-v-6133f14d"]]);var g=h,_=r(779),Z=r(2628),R={name:"Panel",components:{TorrentView:_.Z,Header:g},mixins:[Z.Z],props:{pluginName:{type:String,required:!0}},methods:{download:function(e){var n=this;return(0,c.Z)(regeneratorRuntime.mark((function r(){return regeneratorRuntime.wrap((function(r){while(1)switch(r.prev=r.next){case 0:return r.next=2,n.request("".concat(n.pluginName,".download"),{torrent:e});case 2:case"end":return r.stop()}}),r)})))()}}};const b=(0,w.Z)(R,[["render",u],["__scopeId","data-v-1502d8a8"]]);var U=b}}]);
//# sourceMappingURL=2820-legacy.6e4e2d10.js.map