"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[4826],{4826:function(n,t,e){e.r(t),e.d(t,{default:function(){return l}});var i=e(6252),o={class:"plugin"};function r(n,t,e,r,s,a){var c=(0,i.up)("Loading");return(0,i.wg)(),(0,i.iD)("div",o,[s.loading?((0,i.wg)(),(0,i.j4)(c,{key:0})):s.component?((0,i.wg)(),(0,i.j4)((0,i.LL)(s.component),{key:1,config:s.config},null,8,["config"])):(0,i.kq)("",!0)])}var s=e(8534),a=(e(5666),e(9600),e(1249),e(4916),e(3123),e(7042),e(1539),e(8783),e(3948),e(2628)),c=e(1232),u={name:"Plugin",components:{Loading:c.Z},mixins:[a.Z],props:{pluginName:{type:String,required:!0}},data:function(){return{loading:!1,component:null,config:{}}},computed:{componentName:function(){return this.pluginName.split(".").map((function(n){return n[0].toUpperCase()+n.slice(1)})).join("")}},methods:{refresh:function(){var n=(0,s.Z)(regeneratorRuntime.mark((function n(){var t,o=this;return regeneratorRuntime.wrap((function(n){while(1)switch(n.prev=n.next){case 0:return this.loading=!0,n.prev=1,this.component=(0,i.RC)((function(){return e(3379)("./".concat(o.componentName,"/Index"))})),this.$options.components[this.componentName]=this.component,n.next=6,this.request("config.get_plugins");case 6:if(n.t2=t=n.sent,n.t1=null===n.t2,n.t1){n.next=10;break}n.t1=void 0===t;case 10:if(!n.t1){n.next=14;break}n.t3=void 0,n.next=15;break;case 14:n.t3=t[this.pluginName];case 15:if(n.t0=n.t3,n.t0){n.next=18;break}n.t0={};case 18:this.config=n.t0;case 19:return n.prev=19,this.loading=!1,n.finish(19);case 22:case"end":return n.stop()}}),n,this,[[1,,19,22]])})));function t(){return n.apply(this,arguments)}return t}()},mounted:function(){this.refresh()}},p=e(3744);const f=(0,p.Z)(u,[["render",r],["__scopeId","data-v-3565b88b"]]);var l=f}}]);
//# sourceMappingURL=4826-legacy.580eecfe.js.map