"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[5895],{5895:function(e,n,a){a.r(n),a.d(n,{default:function(){return i}});var u=a(6252);function t(e,n,a,t,p,r){const s=(0,u.up)("Media");return(0,u.wg)(),(0,u.j4)(s,{"plugin-name":"media.mpv"})}var p=a(4196),r={name:"MediaMpv",components:{Media:p["default"]}},s=a(3744);const c=(0,s.Z)(r,[["render",t]]);var i=c}}]);
//# sourceMappingURL=5895.6d58227e.js.map